#include <stdio.h>
#include <stdlib.h>
#include <SDL.h>
#include <gnome.h>
#include <gdk/gdkx.h>
#include "g_main.h"
#include "g_ddraw.h"
#include "g_dsound.h"
#include "g_input.h"
#include "gens.h"
#include "mem_m68k.h"
#include "blit.h"
#include "vdp_io.h"
#include "vdp_rend.h"
#include "misc.h"

int Message_Style;
int FPS_Style;
int Blit_Soft;
int Stretch;
int W_VSync;
int FS_VSync;
int Sleep_Time;
int Effect_Color = 7;
int Flag_Clr_Scr = 0;
clock_t Last_Time = 0, New_Time = 0;
clock_t Used_Time = 0;

static SDL_Surface *screen;

static char Info_String[1024] = "";
static char osd_message[1024] = "";
static int osd_active = 0;
static unsigned int osd_time = 0;

void Put_Info(char *Message, int Duree)
{
	if (Show_Message) {
		strcpy(osd_message, Message);
		osd_time = GetTickCount() + Duree;
		osd_active = 1;
	}
}

int GetTickCount(void)
{
	return SDL_GetTicks();
}

int Flip(void)
{
	int Dep, i;
	static float FPS = 0.0f, frames[8] = {0, 0, 0, 0, 0, 0, 0, 0};
	static unsigned int old_time = 0, view_fps = 0, index_fps = 0;
	unsigned int new_time[2];
	
	if (osd_active) {
		if (GetTickCount() > osd_time) {
			osd_active = 0;
			osd_message[0] = '\0';
		} else Print_Text(osd_message, strlen(osd_message), 10, 210, Message_Style);
	} else if (Show_FPS) {
		/* Screw this performance counter crap, I could give a shit! */
		if (++view_fps >= 10) {
			new_time[0] = GetTickCount();
			
			if (new_time[0] != old_time) frames[index_fps] = 10000 / (float)(new_time[0] - old_time);
			else frames[index_fps] = 2000;
			
			index_fps++; index_fps &= 7;
			FPS = 0.0f;
			
			for (i = 0; i < 8; i++) FPS += frames[i];
			
			FPS /= 8.0f;
			old_time = new_time[0];
			view_fps = 0;
		}
		sprintf(Info_String, "%.1f", FPS);
		Print_Text(Info_String, strlen(Info_String), 10, 210, FPS_Style);
	}
	
	if ((VDP_Reg.Set4 & 0x1) || (Debug)) {
		Dep = 0;
	} else {
		Dep = 64;
	}
	
	SDL_LockSurface(screen);
	Blit((unsigned char*) screen->pixels + (((screen->w*2) * ((240 - VDP_Num_Vis_Lines) >> 1) + Dep) << 1), screen->w*2, 320 - Dep, VDP_Num_Vis_Lines, 32 + Dep * 2);
	SDL_UnlockSurface(screen);
	SDL_UpdateRect(screen, 0, 0, 0, 0);
	return(1);
}

void Update_Emulation(void)
{
	static int Over_Time = 0;
	int current_div;
	
	if (Frame_Skip != -1) {
		if (Sound_Enable) {
			WP = (WP + 1) & (Sound_Segs - 1);
			if (WP == Get_Current_Seg())
				WP = (WP + Sound_Segs - 1) & (Sound_Segs - 1);
			Write_Sound_Buffer(NULL);
		}
		Update_Controllers();
		if (Frame_Number++ < Frame_Skip) {
			Update_Frame_Fast();
		} else {
			Frame_Number = 0;
			Update_Frame();
			Flip();
		}
	} else {
		if (Sound_Enable) {
			while (WP == Get_Current_Seg());
			RP = Get_Current_Seg();
			while (WP != RP) {
				Write_Sound_Buffer(NULL);
				WP = (WP + 1) & (Sound_Segs - 1);
				Update_Controllers();
				if (WP != RP) {
					Update_Frame_Fast();
				} else {
					Update_Frame();
					Flip();
				}
			}
		} else {
			if (CPU_Mode) current_div = 20;
			else current_div = 16 + (Over_Time ^= 1);
			
			New_Time = GetTickCount();
			Used_Time += (New_Time - Last_Time);
			Frame_Number = Used_Time / current_div;
			Used_Time %= current_div;
			Last_Time = New_Time;
			
			if (Frame_Number > 8) Frame_Number = 8;
			
			for (; Frame_Number > 1; Frame_Number--) {
				Update_Controllers();
				Update_Frame_Fast();
			}
			
			if (Frame_Number) {
				Update_Controllers();
				Update_Frame();
				Flip();
			}
		}
	}
}

void Init_Display(void)
{
	char envstring[32];
	
	if (Full_Screen) {
		SDL_QuitSubSystem(SDL_INIT_VIDEO);
		unsetenv("SDL_WINDOWID");
		if (SDL_Init(SDL_INIT_VIDEO)) {
			fprintf(stderr, _("Haha, I knew this wouldn't work!"));
			exit(1);
		}
		if (Render < 1) {
			screen = SDL_SetVideoMode(320, 240, 16, SDL_FULLSCREEN /* 0 */);
		} else {
			screen = SDL_SetVideoMode(640, 480, 16, SDL_FULLSCREEN /* 0 */);
		}
	} else {
		SDL_QuitSubSystem(SDL_INIT_VIDEO);
		if (Render < 1) {
			gtk_widget_set_usize(sockwidget, 336, 240);
		} else {
			gtk_widget_set_usize(sockwidget, 640, 480);
		}
		gtk_widget_queue_resize(sockwidget);
		gtk_widget_realize(sockwidget);
		g_snprintf(envstring, 32, "SDL_WINDOWID=%ld",
			GDK_WINDOW_XWINDOW(sockwidget->window));
		putenv(envstring);
		if (SDL_Init(SDL_INIT_VIDEO)) {
			fprintf(stderr, _("\"Uhhhh!\" --Gefo\n"));
			exit(1);
		}
		if (Render < 1) {
			screen = SDL_SetVideoMode(336, 240, 16, 0);
		} else {
			screen = SDL_SetVideoMode(640, 480, 16, 0);
		}
	}
}

void Update_Crazy_Effect(void)
{
	int i, j, offset;
	int r=0, g=0, b=0, prev_l, prev_p;
	int RB, G;

	for (offset = 336 * 240, j = 0; j < 240; j++) {
		for (i = 0; i < 336; i++, offset--) {
			prev_l = MD_Screen[offset - 336];
			prev_p = MD_Screen[offset - 1];
                      
			if (Mode_555 & 1) {
				RB = ((prev_l & 0x7C1F) + (prev_p & 0x7C1F)) >> 1;
				G = ((prev_l & 0x03E0) + (prev_p & 0x03E0)) >> 1;
				if (Effect_Color & 0x4) {
					r = RB & 0x7C00;
					if (rand() > 0x2C000000) r += 0x0400;
					else r -= 0x0400;
					if (r > 0x7C00) r = 0x7C00;
					else if (r < 0x0400) r = 0;
				}
				if (Effect_Color & 0x2) {
					g = G & 0x03E0;
					if (rand() > 0x2C000000) g += 0x0020;
					else g -= 0x0020;
					if (g > 0x03E0) g = 0x03E0;
					else if (g < 0x0020) g = 0;
				}
				if (Effect_Color & 0x1) {
					b = RB & 0x001F;
					if (rand() > 0x2C000000) b++;
					else b--;
					if (b > 0x1F) b = 0x1F;
					else if (b < 0) b = 0;
				}
			} else {
				RB = ((prev_l & 0xF81F) + (prev_p & 0xF81F)) >> 1;
				G = ((prev_l & 0x07C0) + (prev_p & 0x07C0)) >> 1;
				if (Effect_Color & 0x4) {
					r = RB & 0xF800;
					if (rand() > 0x2C000000) r += 0x0800;
					else r -= 0x0800;
					if (r > 0xF800) r = 0xF800;
					else if (r < 0x0800) r = 0;
				}
				if (Effect_Color & 0x2) {
					g = G & 0x07C0;
					if (rand() > 0x2C000000) g += 0x0040;
					else g -= 0x0040;
					if (g > 0x07C0) g = 0x07C0;
					else if (g < 0x0040) g = 0;
				}
				if (Effect_Color & 0x1) {
					b = RB & 0x001F;
					if (rand() > 0x2C000000) b++;
					else b--;
					if (b > 0x1F) b = 0x1F;
					else if (b < 0) b = 0;
				}
			}
			MD_Screen[offset] = r + g + b;
			}

	}
	Flip();
}

int Toggle_FS(void)
{
	Full_Screen = Full_Screen ? 0 : 1;
	Init_Display();
	
	return 1;
}

void (*Blit)(unsigned char *Dest, int pitch, int x, int y, int offset);
int (*Update_Frame)();
int (*Update_Frame_Fast)();
