#include <stdio.h>
#include <gnome.h>
#include <unistd.h>
#include <pthread.h>
#include <SDL.h>
#include "gens.h"
#include "g_ddraw.h"
#include "g_dsound.h"
#include "g_main.h"
#include "g_input.h"
#include "save.h"
#include "z80.h"
#include "cpu_sh2.h"
#include "cpu_68k.h"
#include "cpu_z80.h"
#include "ym2612.h"
#include "psg.h"
#include "pwm.h"
#include "vdp_io.h"
#include "misc.h"
#include "blit.h"
#include <gdk/gdkx.h>
#include <X11/Xlib.h>
#include "gdkevxlate.h"
#include "gui.h"

GtkWidget *app, *sockwidget, *statusbar, *popmenu;
GnomeUIInfo graphics_menu[], top_menu[];
Display *disp;

static void loads3k_cb(GtkWidget* widget, gpointer foo)
{
	Pre_Load_Rom("/c/kega/s3k.bin");
}

static void exit_cb(GtkWidget* foo, gpointer bar)
{
	gtk_main_quit();
}

static void getrom_cb(GtkWidget* foo, gpointer *bar)
{
	Paused=1;
	Get_Rom();
	Paused=0;
}

static void freerom_cb(GtkWidget* foo, gpointer *bar)
{
	Free_Rom(Game);
}

static void gamegenie_cb(GtkWidget* foo, gpointer *bar)
{
	gui_gamegenie();
}

GnomeUIInfo file_menu[] = {
	GNOMEUIINFO_ITEM_NONE(N_("Open S3K"),
			N_("Open Sonic 3 & Knuckles (debug option)"),
			loads3k_cb),
	GNOMEUIINFO_MENU_OPEN_ITEM(getrom_cb, NULL),
	GNOMEUIINFO_MENU_CLOSE_ITEM(freerom_cb, NULL),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_ITEM_NONE(N_("_Game Genie..."),
			N_("Enter Game Genie codes"),
			gamegenie_cb),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_EXIT_ITEM(exit_cb, NULL),
	GNOMEUIINFO_END
};

static void fullscreen_cb(GtkWidget* foo, gpointer *bar)
{
	int i=0;
	
	if (!Toggle_FS()) {
		Put_Info(_("FS mode switch failed!"), 1500);
		return;
	}
	if (Full_Screen) {
		Put_Info(_("Full screen mode selected"), 1500);
		while(top_menu[i].type) {
			/* g_object_ref(topmenu[i].widget); */
			gtk_container_remove(
				GTK_CONTAINER(GNOME_APP(app)->menubar),
				top_menu[i].widget);
			gtk_menu_append(GTK_MENU(popmenu), top_menu[i].widget);
			/* g_object_unref(top_menu[i].widget); */
			i++;
		}
	} else {
		Put_Info(_("Window mode selected"), 1500);
		while(top_menu[i].type) {
			/* g_object_ref(topmenu[i].widget); */
			gtk_container_remove(GTK_CONTAINER(popmenu), top_menu[i].widget);
			gtk_menu_bar_append(
				GTK_MENU_BAR(GNOME_APP(app)->menubar),
				top_menu[i].widget);
			/* g_object_unref(top_menu[i].widget); */
			i++;
		}
	}
}

static void render_cb(GtkWidget* foo, gpointer *param)
{
	int i;
	
	i = GPOINTER_TO_INT(param);
	Render = i;
	switch (Render) {
		case 0:	Blit = Have_MMX ? Blit_X1_MMX : Blit_X1;
			Put_Info(_("Render selected : NORMAL"), 1500);
			break;
		case 1: Blit = Have_MMX ? Blit_X2_MMX : Blit_X2;
			Put_Info(_("Render selected : DOUBLE"), 1500);
			break;
		case 2: Blit = Have_MMX ? Blit_X2_Int_MMX : Blit_X2_Int;
			Put_Info(_("Render selected : INTERPOLATED"), 1500);
			break;
		case 3: Blit = Have_MMX ? Blit_Scanline_MMX : Blit_Scanline;
			Put_Info(_("Render selected : FULL SCANLINE"), 1500);
			break;
		case 4: if (Have_MMX) {
				Blit = Blit_Scanline_50_MMX;
				Put_Info(_("Render selected : 50% SCANLINE"), 1500);
			} else {
				Render = 3;
				Blit = Blit_Scanline;
				Put_Info(_("Render selected : FULL SCANLINE"), 1500);
			}
			break;
		case 5:	if (Have_MMX) {
				Blit = Blit_Scanline_25_MMX;
				Put_Info(_("Render selected : 25% SCANLINE"), 1500);
			} else {
				Render = 3;
				Blit = Blit_Scanline;
				Put_Info(_("Render selected : FULL SCANLINE"), 1500);
			}
			break;
		case 6: Blit = Have_MMX ? Blit_Scanline_Int_MMX : Blit_Scanline_Int;
			Put_Info(_("Render selected : INTERPOLATED SCANLINE"), 1500);
			break;
		case 7: if (Have_MMX) {
				Blit = Blit_Scanline_50_Int_MMX;
				Put_Info(_("Render selected : INTERPOLATED 50% SCANLINE"), 1500);
			} else {
				Render = 6;
				Blit = Blit_Scanline_Int;
				Put_Info(_("Render selected : INTERPOLATED SCANLINE"), 1500);
			}
			break;
		case 8:	if (Have_MMX) {
				Blit = Blit_Scanline_25_Int_MMX;
				Put_Info(_("Render selected : INTERPOLATED 25% SCANLINE"), 1500);
			} else {
				Render = 6;
				Blit = Blit_Scanline_Int;
				Put_Info(_("Render selected : INTERPOLATED SCANLINE"), 1500);
			}
			break;
		case 9:	if (Have_MMX) {
				Blit = Blit_2xSAI_MMX;
				Put_Info(_("Render selected : 2XSAI KREED'S ENGINE"), 1500);
			} else {
				Render = 6;
				Blit = Blit_Scanline_Int;
				Put_Info(_("Render selected : INTERPOLATED SCANLINE"), 1500);
			}
			break;
	}
	if (sockwidget) Init_Display();
}

GnomeUIInfo render_radio_list[] = {
	{ GNOME_APP_UI_ITEM, N_("_Normal"),
	  N_("Select no post-processing on frames (fastest)"),
	  render_cb, GINT_TO_POINTER(0), NULL, GNOME_APP_PIXMAP_NONE, NULL,
	  0, 0, NULL},
	{ GNOME_APP_UI_ITEM, N_("_Double"),
	  N_("Scale up by a factor of two"),
	  render_cb, GINT_TO_POINTER(1), NULL, GNOME_APP_PIXMAP_NONE, NULL,
	  0, 0, NULL},
	{ GNOME_APP_UI_ITEM, N_("_Interpolated"),
	  N_("Double, and blend in-between pixels"),
	  render_cb, GINT_TO_POINTER(2), NULL, GNOME_APP_PIXMAP_NONE, NULL,
	  0, 0, NULL},
	{ GNOME_APP_UI_ITEM, N_("_Scanline"),
	  N_("Double, inserting black rows"),
	  render_cb, GINT_TO_POINTER(3), NULL, GNOME_APP_PIXMAP_NONE, NULL,
	  0, 0, NULL},
	{ GNOME_APP_UI_ITEM, N_("_50% Scanline"),
	  N_("Double, inserting half-bright rows"),
	  render_cb, GINT_TO_POINTER(4), NULL, GNOME_APP_PIXMAP_NONE, NULL,
	  0, 0, NULL},
	{ GNOME_APP_UI_ITEM, N_("_25% Scanline"),
	  N_("Double, inserting quarter-bright rows"),
	  render_cb, GINT_TO_POINTER(5), NULL, GNOME_APP_PIXMAP_NONE, NULL,
	  0, 0, NULL},
	{ GNOME_APP_UI_ITEM, N_("Interpolated Scanline"),
	  N_("Double, interpolate and insert scanlines (TV Mode)"),
	  render_cb, GINT_TO_POINTER(6), NULL, GNOME_APP_PIXMAP_NONE, NULL,
	  0, 0, NULL},
	{ GNOME_APP_UI_ITEM, N_("Interpolated 50% Scanline"),
	  N_("Double, interpolate and insert 50% scanlines"),
	  render_cb, GINT_TO_POINTER(7), NULL, GNOME_APP_PIXMAP_NONE, NULL,
	  0, 0, NULL},
	{ GNOME_APP_UI_ITEM, N_("Interpolated 25% Scanline"),
	  N_("Double, interpolate and insert 25% scanlines"),
	  render_cb, GINT_TO_POINTER(8), NULL, GNOME_APP_PIXMAP_NONE, NULL,
	  0, 0, NULL},
	{ GNOME_APP_UI_ITEM, N_("2xSAI (_Kreed)"),
	  N_("Use Kreed's amazing Double, Scale and Interpolate filter"),
	  render_cb, GINT_TO_POINTER(9), NULL, GNOME_APP_PIXMAP_NONE, NULL,
	  0, 0, NULL},
	GNOMEUIINFO_END
};

GnomeUIInfo graphics_render_menu[] = {
	GNOMEUIINFO_RADIOLIST(render_radio_list),
	GNOMEUIINFO_END
};

GnomeUIInfo graphics_menu[] = {
	{GNOME_APP_UI_TOGGLEITEM,
			N_("_Full Screen"),
			N_("Switch to full screen mode"),
			fullscreen_cb,
			NULL,
			NULL,
			GNOME_APP_PIXMAP_NONE,
			NULL,
			GDK_Return,
			GDK_MOD1_MASK,
			NULL
	},
	GNOMEUIINFO_SUBTREE(N_("_Render"), graphics_render_menu),
	GNOMEUIINFO_END
};

GnomeUIInfo cpu_menu[] = {
	GNOMEUIINFO_END
};

GnomeUIInfo sound_menu[] = {
	GNOMEUIINFO_END
};

GnomeUIInfo option_menu[] = {
	GNOMEUIINFO_END
};

GnomeUIInfo help_menu[] = {
#if 0
	GNOMEUIINFO_HELP("gnome-gens"),
#endif
	GNOMEUIINFO_MENU_ABOUT_ITEM(about, NULL),
	GNOMEUIINFO_END
};

GnomeUIInfo top_menu[] = {
	GNOMEUIINFO_MENU_FILE_TREE(file_menu),
	GNOMEUIINFO_SUBTREE(N_("_Graphics"), graphics_menu),
	GNOMEUIINFO_SUBTREE(N_("_CPU"), cpu_menu),
	GNOMEUIINFO_SUBTREE(N_("_Sound"), sound_menu),
	GNOMEUIINFO_SUBTREE(N_("_Option"), option_menu),
	GNOMEUIINFO_MENU_HELP_TREE(help_menu),
	GNOMEUIINFO_END
};

int Build_Main_Menu(void)
{
	render_cb(NULL, GINT_TO_POINTER(Render));
	gtk_check_menu_item_set_active(
		GTK_CHECK_MENU_ITEM(render_radio_list[Render].widget), TRUE);
	/* I know the manual says absolutely not to do this, but I just need to set
	   the damn flag -- I'm not trying to use the goddam callback! That happens
	   later! */
	GTK_CHECK_MENU_ITEM(graphics_menu[0].widget)->active=Full_Screen;
	if (!Have_MMX) {
		/* Disable renderers that are only available for MMX */
		gtk_widget_set_sensitive(GTK_WIDGET(render_radio_list[4].widget), FALSE);
		gtk_widget_set_sensitive(GTK_WIDGET(render_radio_list[5].widget), FALSE);
		gtk_widget_set_sensitive(GTK_WIDGET(render_radio_list[7].widget), FALSE);
		gtk_widget_set_sensitive(GTK_WIDGET(render_radio_list[8].widget), FALSE);
		gtk_widget_set_sensitive(GTK_WIDGET(render_radio_list[9].widget), FALSE);
	}
	return 0;
}

static gint delete_event_cb(GtkWidget* w, GdkEventAny* e, gpointer foo)
{
	gtk_main_quit();
	return FALSE;
}

#ifdef USING_THREADS
void* emulator_thread(void *foo)
#else
gint emulator_thread(void *foo)
#endif
{
#ifdef USING_THREADS
	for(;;) {
#endif
	SDL_Event event;
	XEvent xevent;
	GdkEvent gevent;
	
	memset(&xevent, 0, sizeof(xevent));
	
	if (Full_Screen) {
		while(SDL_PollEvent(&event)) {
			switch(event.type) {
			    case SDL_KEYDOWN:
			    case SDL_KEYUP:
			        xevent.type = 
			            (event.key.state == SDL_PRESSED) ?
			            KeyPress : KeyRelease;
					xevent.xkey.send_event = TRUE;
			        xevent.xkey.display = disp;
			        xevent.xkey.window = GDK_WINDOW_XWINDOW(app->window);
			        xevent.xkey.root = GDK_ROOT_WINDOW();
			        xevent.xkey.time = GDK_CURRENT_TIME;
			        xevent.xkey.same_screen = TRUE;
				    switch(event.key.keysym.sym) {
						case 0x08:
						case 0x09:
						case 0x0A:
						case 0x0B:
						case 0x0D:
						case 0x13:
						case 0x14:
						case 0x15:
						case 0x1B:
						case 0xFF:
							event.key.keysym.sym |= 0xFF00;
							break;
						default:
							break;
					}
			    	xevent.xkey.keycode = 
						XKeysymToKeycode(disp, event.key.keysym.sym);
			    	xevent.xkey.state = 
			    	    ((event.key.keysym.mod & KMOD_SHIFT) ?
			    	    	ShiftMask : 0) |
			    	    ((event.key.keysym.mod & KMOD_CTRL) ?
			    	    	ControlMask : 0) |
			    	    ((event.key.keysym.mod & KMOD_CAPS) ?
			    	    	LockMask : 0) |
			    	    ((event.key.keysym.mod & KMOD_ALT) ?
			    	    	Mod1Mask : 0);
			    	printf("Sending: %s 0x%x 0x%x from 0x%x 0x%x\n",
			    	    (xevent.type == KeyPress) ?
			    	    "press" : "release",
			    	    xevent.xkey.state,
			    	    xevent.xkey.keycode,
						event.key.keysym.mod,
						event.key.keysym.sym);
					gdk_event_translate(&gevent, &xevent);
					gdk_event_put(&gevent);
					break;
				case SDL_MOUSEBUTTONDOWN:
					switch (event.button.button) {
						case SDL_BUTTON_RIGHT:
							SDL_WM_GrabInput(SDL_GRAB_OFF);
							gtk_menu_popup(
								GTK_MENU(popmenu), 
								NULL, 
								NULL, 
								NULL,
								NULL,
								3,
								SDL_GetTicks());
							break;
						case SDL_BUTTON_LEFT:
							SDL_WM_GrabInput(SDL_GRAB_ON);
							gtk_menu_popdown(GTK_MENU(popmenu));
							break;
						default:
							break;
					}
					break;
				default:
			    	break;
			}
		}
	}

		if (Genesis_Started || _32X_Started || SegaCD_Started) {
			if ((!Paused)) {
				Update_Emulation();
			} else {
				usleep(100000);
			}
		} else {
			Update_Crazy_Effect();
			usleep(1000);
		}
#ifdef USING_THREADS
	}
#else
	return TRUE;
#endif
}

GdkColor black = { 0, 0, 0, 0 };

void init(int argc, char** argv)
{
	GtkWidget *box;
	GtkStyle *style;
	/* FIXME: Check for DISPLAY before running gnome_init */
	/* FIXME: Come up with some command-line options and use popt */
	
	disp = XOpenDisplay(getenv("DISPLAY"));

	Identify_CPU();

	/* Unfortunately necessary. Gnome's crash dialog nicely steps on SDL's segv
	   handler, so if the program crashes in fullscreen mode, the display will
	   be locked to 640x480... */
	putenv("GNOME_DISABLE_CRASH_DIALOG=1");
	gnome_init("gnome-gens", "0.0", argc, argv);
	
	Load_Config(NULL, NULL);
	
	app = gnome_app_new("gnome-gens", _("Gens - Idle"));
	gtk_widget_add_events(app, GDK_KEY_RELEASE_MASK);
	
	statusbar = gnome_appbar_new(FALSE, TRUE, GNOME_PREFERENCES_USER);
	gnome_app_set_statusbar(GNOME_APP(app), statusbar);

	gnome_app_create_menus(GNOME_APP(app), top_menu);
	popmenu = gtk_menu_new();
	Build_Main_Menu();
	
	gnome_app_install_menu_hints(GNOME_APP(app), top_menu);
	
	box = gtk_alignment_new(0.5, 0.5, 0, 0);
	style = gtk_widget_get_style(app);
	style->bg[GTK_STATE_NORMAL] = black;
	gtk_widget_set_style(box, style);
	sockwidget = gtk_socket_new();
	gnome_app_set_contents(GNOME_APP(app), box);
	gtk_container_add(GTK_CONTAINER(box), sockwidget);
	
	gtk_window_set_policy(GTK_WINDOW(app), FALSE, FALSE, TRUE);

	gtk_signal_connect(GTK_OBJECT(app), "delete_event",
			GTK_SIGNAL_FUNC(delete_event_cb), NULL);
	
	gtk_widget_show_all(app);
	
	Init_Display();
	Init_Input();
	
	MSH2_Init();
	SSH2_Init();
	M68K_Init();
	S68K_Init();
	Z80_Init();
	
	YM2612_Init(CLOCK_NTSC / 7, Sound_Rate, YM2612_Improv);
	PSG_Init(CLOCK_NTSC / 15, Sound_Rate);
	PWM_Init();
}

int main(int argc, char** argv)
{
#ifdef USING_THREADS
	pthread_t thread;
#endif
	
	init(argc, argv);
	
	Init_Genesis_Bios();	
	
#ifdef USING_THREADS
	if(pthread_create(&thread, NULL, emulator_thread, NULL)) {
		fprintf(stderr, _("*sob*"));
		exit(1);
	}
#else
	gtk_idle_add(emulator_thread, NULL);
#endif

	gtk_main();

	Save_Config(NULL);

	return 0;
}

char Str_Tmp[1024];
char Gens_Path[1024];

struct Rom *Game = NULL;

int Auto_Pause = 0;
int Auto_Fix_CS = 0;
int Show_LED = 0;
int Show_Message = 0;
int Show_FPS = 0;
int Fast_Blur = 0;
int SegaCD_Accurate = 0;
int Country_Order[3];
int Country = -1;
int Render = 0;
int Full_Screen = -1;
int Gens_Priority;
int Intro_Style = 0;
int Language = 0;
int Net_Play = 0;
int Paused = 0;
int Kaillera_Client_Running = 0;
